// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::{AluBranchMOp, UnitOutRegNum},
    unit::{
        unit_base::{unit_base, UnitToRegAlloc},
        DynUnit, DynUnitWrapper, UnitKind, UnitMOp, UnitTrait,
    },
};
use fayalite::{
    intern::{Intern, Interned},
    prelude::*,
};

#[hdl_module]
pub fn alu_branch(config: &CpuConfig, unit_index: usize) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let unit_to_reg_alloc: UnitToRegAlloc<
        AluBranchMOp<UnitOutRegNum<DynSize>, DynSize>,
        (),
        DynSize,
        DynSize,
        DynSize,
    > = m.output(config.unit_to_reg_alloc(
        AluBranchMOp[config.unit_out_reg_num()][config.p_reg_num_width()],
        (),
    ));
    #[hdl]
    let unit_base = instance(unit_base(
        config,
        unit_index,
        Expr::ty(unit_to_reg_alloc).input_insn.data.HdlSome,
        (),
    ));
    connect(unit_to_reg_alloc, unit_base.unit_to_reg_alloc);
    connect(unit_base.cd, cd);
    connect(unit_base.execute_start.ready, true); // TODO: finish
    connect(
        unit_base.execute_end,
        Expr::ty(unit_base.execute_end).HdlNone(),
    ); // TODO: finish
}

#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
pub struct AluBranch {
    config: Interned<CpuConfig>,
    module: Interned<Module<alu_branch>>,
}

impl AluBranch {
    pub fn new(config: &CpuConfig, unit_index: usize) -> Self {
        Self {
            config: config.intern(),
            module: alu_branch(config, unit_index),
        }
    }
}

impl UnitTrait for AluBranch {
    type Type = alu_branch;
    type ExtraOut = ();
    type MOp = AluBranchMOp<UnitOutRegNum<DynSize>, DynSize>;

    fn ty(&self) -> Self::Type {
        self.module.io_ty()
    }

    fn extra_out_ty(&self) -> Self::ExtraOut {
        ()
    }

    fn mop_ty(&self) -> Self::MOp {
        self.module.io_ty().unit_to_reg_alloc.mop_ty()
    }

    fn unit_kind(&self) -> UnitKind {
        UnitKind::AluBranch
    }

    fn extract_mop(
        &self,
        mop: Expr<UnitMOp<UnitOutRegNum<DynSize>, DynSize>>,
    ) -> Expr<HdlOption<Self::MOp>> {
        UnitMOp::alu_branch_mop(mop)
    }

    fn module(&self) -> Interned<Module<Self::Type>> {
        self.module
    }

    fn unit_to_reg_alloc(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<UnitToRegAlloc<Self::MOp, Self::ExtraOut, DynSize, DynSize, DynSize>> {
        this.unit_to_reg_alloc
    }

    fn cd(&self, this: Expr<Self::Type>) -> Expr<ClockDomain> {
        this.cd
    }

    fn to_dyn(&self) -> DynUnit {
        DynUnitWrapper(*self).to_dyn()
    }
}
